"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.P = function (m, n) {
    n % 1 !== 0 && (n = Math.floor(n));
    m % 1 !== 0 && (m = Math.floor(m));
    let p = 1;
    while (n--) {
        p *= m--;
    }
    return p;
};
exports.C = function (m, n) {
    if (n > m) {
        return 0;
    }
    return exports.P(m, n) / exports.P(n, n);
};
class AbstractIndexes {
    constructor(m, indexes) {
        this.m = m;
        this.indexes = indexes;
    }
    pickFrom(ary) {
        if (this.m > ary.length) {
            throw new RangeError;
        }
        let result = [];
        this.forIndex(this.indexes, (e, i) => e && (result[result.length] = ary[i]));
        return result;
    }
    omitFrom(ary) {
        if (this.m > ary.length) {
            throw new RangeError;
        }
        let result = [];
        this.forIndex(this.negateIndexes(), (e, i) => e && (result[result.length] = ary[i]));
        return result;
    }
}
class AbstractCombination {
    constructor(m, r) {
        this.m = m;
        this.r = r;
        this.count = 0;
        this.count = exports.C(m, r);
    }
    isValid() {
        return 0 < this.r && this.m >= this.r;
    }
    [Symbol.iterator]() {
        return this;
    }
}
class Indexes30 extends AbstractIndexes {
    constructor(m, indexes) {
        super(m, indexes);
    }
    negateIndexes() {
        const max = (1 << this.m) - 1;
        return max ^ this.indexes;
    }
    forIndex(indexes, callback) {
        let i = 0;
        for (let n = indexes; n; n >>>= 1, i++) {
            callback(Boolean(n & 1), i);
        }
    }
}
class Combination30 extends AbstractCombination {
    constructor(m, r) {
        super(m, r);
        this.maxIndex = 1 << m;
        this.indexes = this.isValid() ? (1 << r) - 1 : this.maxIndex;
    }
    next() {
        const done = this.indexes >= this.maxIndex;
        const picker = new Indexes30(this.m, done ? 0 : this.indexes);
        this.indexes = Combination30.nextIndex_30bit(this.indexes);
        return { done, value: picker };
    }
    isValid() {
        return super.isValid() && 31 > this.m;
    }
}
Combination30.nextIndex_30bit = function (ind) {
    var smallest = ind & -ind, ripple = ind + smallest, newSmallest = ripple & -ripple, ones = ((newSmallest / smallest) >> 1) - 1;
    return ripple | ones;
};
exports.Combination30 = Combination30;
class IndexesBig extends AbstractIndexes {
    constructor(m, indexes) {
        super(m, indexes);
    }
    negateIndexes() {
        let preCount = this.m - this.indexes.length;
        const negate = this.indexes.map(i => +!i);
        return negate.concat(Array(preCount).fill(1));
    }
    forIndex(indexes, callback) {
        const length = indexes.length;
        for (let i = 0; i < length; i++) {
            callback(Boolean(indexes[i]), i);
        }
    }
}
class CombinationBig extends AbstractCombination {
    constructor(m, r) {
        super(m, r);
        this.indexes = this.isValid() ? Array(r).fill(1) : Array(m + 1);
    }
    next() {
        const done = this.m < this.indexes.length;
        const index = new IndexesBig(this.m, done ? [] : this.indexes.slice());
        this.indexes = CombinationBig.nextIndex_30bit(this.indexes, this.r);
        return { done, value: index };
    }
}
CombinationBig.nextIndex_30bit = function (indexes, nelem) {
    let result = indexes;
    let j = nelem;
    for (let i = result.length - 1; i >= 0; i--) {
        if (result[i] == 1) {
            j--;
        }
        else {
            break;
        }
    }
    if (j == 0) {
        result[result.length] = 1;
        for (let k = result.length - 2; k >= 0; k--) {
            result[k] = (k < nelem - 1) ? 1 : 0;
        }
    }
    else {
        let i1 = -1;
        let i0 = -1;
        for (let i = 0; i < result.length; i++) {
            if (result[i] == 0 && i1 != -1) {
                i0 = i;
            }
            else if (result[i] == 1) {
                i1 = i;
            }
            if (i0 != -1 && i1 != -1) {
                result[i0] = 1;
                result[i1] = 0;
                break;
            }
        }
        j = nelem;
        for (let i = result.length - 1; i >= i1; i--) {
            if (result[i] == 1) {
                j--;
            }
        }
        for (let i = 0; i < i1; i++) {
            result[i] = i < j ? 1 : 0;
        }
    }
    return result;
};
exports.CombinationBig = CombinationBig;
exports.Combination = function (m, r) {
    return 31 > m ? new Combination30(m, r) : new CombinationBig(m, r);
};
//# sourceMappingURL=combination.js.map